'''
C-Language Backend Code for Simulating Room Impulse Response

author: github.com/ludlows
2019-11 

This program is designed with the hope that it will be useful, but WITHOUT ANY GUARANTEE.
'''
# expose the rir function
from .cyrir import rir



'''
Room Class for Simulating Room Impulse Response

author: github.com/ludlows
2019-11 

This program is designed with the hope that it will be useful, but WITHOUT ANY GUARANTEE.
'''
import numbers
import numpy as np
from .microphone import Microphone
from .speaker import Speaker

class Room:
    """
    Room class
    Args: 
        size (meter):  (size_x, size_y, size_z) numerical tuple with length of 3
                        size_x (meter): Room Length along X Axis
                        size_y (meter): Room Length along Y Axis
                        size_z (meter): Room Length along Z Axis
        order       : integer value >= -1 (maximal), wall reflection order, optional
        name        : str, optional
    """
    _room_id = 0
    def __init__(self, size, order=-1, name=None):
        """
        Args: 
            size (meter):  (size_x, size_y, size_z) numerical tuple with length of 3
                            size_x (meter): Room Length along X Axis
                            size_y (meter): Room Length along Y Axis
                            size_z (meter): Room Length along Z Axis
            order       : integer value >= -1, 
            name        : str
        """
        if round(order) < -1:
            raise ValueError('Reflection Order Should be an integer greater than -2')
        self._order = round(order)
        if len(size) != 3:
            raise ValueError('The length of Room Size should be 3 !!!')
        if not all([isinstance(v, numbers.Number) for v in size]):
            raise ValueError('All the elements of Room Size should be numeric!!!')
        self._size_x, self._size_y , self._size_z = size
        if not name:
            self._name = "Room_{:d}".format(self._room_id)
        else:
            self._name = name
        self._room_id += 1
        self._mics = None 
        self._speakers = None 

    def _check_pos(self, obj_with_pos):
        """
        Returns True if the position of obj_with_pos fits with the Room Size
        Args:
            obj_with_pos: Microphone or Speaker 
        """ 
        x, y, z = obj_with_pos.get_pos()
        if x < 0 or x > self._size_x or y < 0 or y > self._size_y or z < 0 or z > self._size_z:
            return False
        return True

    def setup_mic_speaker(self, mic_or_mics, speaker_or_speakers):
        """
        Args:
            mics    : Microphone object or list of Microphones
            speakers: Speaker object or list of Speakers

        """
        # setup mic and check types
        if isinstance(mic_or_mics, Microphone):
            self._mics = (mic_or_mics,)
        elif isinstance(mic_or_mics, list) or isinstance(mic_or_mics, tuple):
            mics = list(mic_or_mics)
            if any([not isinstance(mic, Microphone) for mic in mics]):
                raise ValueError('Some Objects are not Microphone Objects')
            if any([not self._check_pos(obj) for obj in mics]):
                raise ValueError('Some Microphone Positions are not compatible with the Rooom Size')
            self._mics = tuple(mics)
        else:
            raise ValueError('The mic_or_mics value should be Microphone object or list of Microphone Objects')

        # setup speakers and check types
        if isinstance(speaker_or_speakers, Speaker):
            self._speakers = (speaker_or_speakers,)
        elif isinstance(speaker_or_speakers, list) or isinstance(speaker_or_speakers, tuple):
            spks = list(speaker_or_speakers)
            if any([not isinstance(spk, Speaker) for spk in spks]):
                raise ValueError('Some Objects are not Spkeaker Objects')
            if any([not self._check_pos(obj) for obj in spks]):
                raise ValueError('Some Speaker Positions are not compatible with the Rooom Size')
            self._speakers = tuple(spks)
        else:
            raise ValueError('The speaker_or_speakers value should be Speaker object or list of Speaker Objects')
    
    def get_reflect_order(self):
        return self._order 

    def get_size(self):
        return (self._size_x, self._size_y, self._size_z) 

    def mic_speaker_combination(self):
        """
        Returns list of tuples like ((mic1, mic2, ...), speaker)
        """
        if (not self._mics) or (not self._speakers):
            raise RuntimeError("The Microphones and Speakers are not setup yet.")
        comb = []
        for spk in self._speakers:
            comb.append((tuple(self._mics), spk))
        return tuple(comb)
        
    def __str__(self):
        return "{}_x_{:.1f}_y_{:.1f}_z_{:.1f}".format(self._name, self._size_x, self._size_y, self._size_z)


class ReverbRoom(Room):
    """
    Reverb Time Room class (Defined by T60) 
    Args:
        size:  (x,y,z) size of 3 dimensions 
        rt60:  reverberation time   (second)
        order: reflection order (default -1), options: 
        name: str
    Reference: rt60: the time required for the intensity of reflected sound rays
                         to be down 60dB from thr direct path sound ray
    """
    def __init__(self, size, rt60, order=-1, name=None):
        """
        Args:
            size:  (x,y,z) size of 3 dimensions 
            rt60:  reverberation time   (second)
            order: reflection order (default -1), options: 
            name: str
        Reference: rt60: the time required for the intensity of reflected sound rays
                         to be down 60dB from thr direct path sound ray
        """
        super(ReverbRoom, self).__init__(size, order, name)
        self._name += "_Reverb"
        if rt60 < 0:
            raise ValueError("The RT60 (in seconds) should be >= 0.")
        self._rt60 = rt60
    
    def get_rt60(self):
        return self._rt60


class ReflectRoom(Room):
    """
    Reflection Coefficient Room class (Defined by Beta)
    Args:
        size:        (x,y,z) size of 3 dimensions 
        beta_array:  reflection_coefficients (betaX1, betaX2, betaY1, betaY2, betaZ1, betaZ2)
                        each value ranges from 0 to 1, [0,1]
        order:       reflection order (default -1 as maximal), optional
        name:        str, optional
    """
    def __init__(self, size, beta_array, order=-1, name=None):
        """
        Args:
            size:        (x,y,z) size of 3 dimensions 
            beta_array:  reflection_coefficients (betaX1, betaX2, betaY1, betaY2, betaZ1, betaZ2)
                         each value ranges from 0 to 1, [0,1]
            order:       reflection order (default -1), options: 
            name:         str
        """
        if not len(beta_array) == 6:
            raise ValueError("The beta array should be with a length of 6.")
        if any([b>1 or b<0 for b in beta_array]):
            raise ValueError('The reflection coefficient should be in the range [0,1].')
        self._beta_arr = tuple(beta_array)
        super(ReflectRoom, self).__init__(size, order, name)
        self._name += "_Reflect"

    def get_beta_array(self):
        return self._beta_arr



'''
Speaker Class for Simulating Room Impulse Response

author: github.com/ludlows
2019-11 

This program is designed with the hope that it will be useful, but WITHOUT ANY GUARANTEE.
'''
import numbers

class Speaker:
    """
    Speaker class
    Args:
            position : (x,y,z) or [x,y,z]
            name     : speaker name, optional
    """
    _speaker_id = 0
    def __init__(self, position, name=None):
        """
        Args:
            position : (x,y,z)
            name     : speaker name
        """
        if len(position) != 3:
            raise ValueError('The length of Speaker position should be 3.')
        if not all([isinstance(v, numbers.Number) for v in position]):
            raise ValueError('All the elements of Speaker position should be numeric.')
        self._pos = tuple(position)
        if not name:
            self._name = "Speaker_{:d}".format(self._speaker_id)
        else:
            self._name = name
        self._speaker_id += 1

    def get_pos(self):
        return self._pos 

    def get_name(self):
        return self._name 

    def __str__(self):
        return "{}_x_{:.1f}_y_{:.1f}_z_{:.1f}".format(self._name, *self._pos)




# distutils: language=C

"""
2019-Dec
github.com/ludlows
Python Wrapper for Room Impulse Response
"""


import cython
import numpy as np
cimport numpy as np


cdef extern from "rir.c":
    cdef void comp_rir(double sound_speed, double fs, 
	          double size_x, double size_y, double size_z, 
	          double  mic_x, double  mic_y, double  mic_z, 
	          double  src_x, double  src_y, double  src_z, 
	          double * beta_arr,                           
	          double mic_azimuth, double mic_elevation,    
	          double * impulse, int impulse_len,           
              int     high_pass,                           
	          char    mic_type,                            
	          int     reflect_order)

cpdef np.ndarray[np.float64_t, ndim=1, mode="c"] rir(double sound_speed, double fs, 
        np.ndarray[np.float64_t, ndim=1, mode="c"] room_size,
        np.ndarray[np.float64_t, ndim=1, mode="c"] mic_pos, 
        np.ndarray[np.float64_t, ndim=1, mode="c"] src_pos,
        np.ndarray[np.float64_t, ndim=1, mode="c"] beta_arr,
        double mic_azimuth,
        double mic_elevation,
        int    impulse_len,
        int    high_pass,
        char   mic_type,
        int    reflect_order):
    
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] impulse = np.zeros((impulse_len,), dtype=np.float64)
    comp_rir(
        sound_speed, fs, 
        room_size[0], room_size[1], room_size[2],
        mic_pos[0], mic_pos[1], mic_pos[2],
        src_pos[0], src_pos[1], src_pos[2],
        &(beta_arr[0]),
        mic_azimuth, mic_elevation,
        &(impulse[0]), impulse_len,
        high_pass,
        mic_type,
        reflect_order)
    return impulse



'''
Microphone Class for Simulating Room Impulse Response

author: github.com/ludlows
2019-11 

This program is designed with the hope that it will be useful, but WITHOUT ANY GUARANTEE.
'''
import numbers
import math

from .speaker import Speaker

class Microphone:
    """
    Base Class for microphones
    Args:
        position: tuple (x,y,z) or list [x,y,z]
    """
    _mic_id = 0
    def __init__(self, position, name=None):
        """
        Args:
            position: tuple (x,y,z) or list [x,y,z]
        """
        if len(position) != 3:
            raise ValueError('The length of microphone position should be 3.')
        if not all([isinstance(v, numbers.Number) for v in position]):
            raise ValueError('All the elements of microphone position should be numeric.')
        self._pos = tuple(position)
        if not name:
            self._name = "Mic_{:d}".format(self._mic_id)
        else:
            self._name = name 
        self._mic_id += 1
    
    def get_pos(self):
        return self._pos

    def generate_speaker(self, radius, azimuth_deg, elevation_deg=0):
        """
        Args:
            radius    (meter)     :  the distance between microphone and speaker, meter
            azimuth_deg   (degree):  regarding the microphone as the center, the azimuth angle (degree) from mic to speaker
            elevation_deg (degree):  regarding the microphone as the center, the elevation angle (degree) from  mic to speaker
        
        Returns:
            Speaker instance
        """
        x = self._pos[0] + radius * math.cos(elevation_deg / 180.0 * math.pi) * math.cos(azimuth_deg / 180.0 * math.pi)
        y = self._pos[1] + radius * math.cos(elevation_deg / 180.0 * math.pi) * math.sin(azimuth_deg / 180.0 * math.pi)
        z = self._pos[2] + radius * math.sin(elevation_deg / 180.0 * math.pi)
        return Speaker((x,y,z))
    
    def __str__(self):
        return "{}_x_{:.1f}_y_{:.1f}_z_{:.1f}".format(self._name, *self._pos)
    



class Omni(Microphone):
    """
    class for omnidirectional microphone
    Args:
        position: [x,y,z] or (x,y,z) in meter
        name    : str, optional
    """
    def __init__(self, position, name=None):
        """
        position: [x,y,z] or (x,y,z) in meter
        name    : str, optional
        """
        super(Omni, self).__init__(position, name)
        self._name += "_Omni"


class Hypercardioid(Microphone):
    """
    class for hypercardioid microphone
    Args:
        position:     [x,y,z] or (x,y,z) in meter
        orientation:  (azimuth_deg, elevation_deg) in degree
        name       :  str, optional
    """
    def __init__(self, position, orientation, name=None):
        """
        Args:
            position:     [x,y,z] or (x,y,z) in meter
            orientation:  (azimuth_deg, elevation_deg) in degree
            name       : str
        """
        super(Hypercardioid, self).__init__(position, name)
        if len(orientation) != 2:
            raise ValueError('The length of microphone orientation should be 2 .')
        self._orient = tuple(orientation)
        self._name += "_Hypercardioid"
    
    def get_orient(self):
        """
        Returns (azimuth_deg, elevation_deg) the orientation of Hypercardioid Beam Pattern 
        """
        return self._orient
    
    def __str__(self):
        return "{}_azimDeg_{:.1f}_elevDeg_{:.1f}".format(super().__str__(), *self._orient) 


class Cardioid(Microphone):
    """
    class for Cardioid microphone
    Args:
        position:     [x,y,z] or (x,y,z) in meter
        orientation:  (azimuth_deg, elevation_deg) in degree
        name       :  str, optional
    """
    def __init__(self, position, orientation, name=None):
        """
        Args:
            position:     [x,y,z] or (x,y,z) in meter
            orientation:  (azimuth_deg, elevation_deg) in degree
            name       : str
        """
        super(Cardioid, self).__init__(position, name)
        if len(orientation) != 2:
            raise ValueError('The length of microphone orientation should be 2 .')   
        self._orient = tuple(orientation)
        self._name += "_Cardioid"
    
    def get_orient(self):
        """
        Returns (azimuth_deg, elevation_deg) the orientation of Hypercardioid Beam Pattern 
        """
        return self._orient
    
    def __str__(self):
        return "{}_azimDeg_{:.1f}_elevDeg_{:.1f}".format(super().__str__(), *self._orient) 

    
class Dipole(Microphone):
    """
    class for Dipole microphone
     Args:
        position:     [x,y,z] or (x,y,z) in meter
        orientation:  (azimuth_deg, elevation_deg) in degree
        name       :  str, optional
    """
    def __init__(self, position, orientation, name=None):
        """
        Args:
            position:     [x,y,z] or (x,y,z) in meter
            orientation:  (azimuth_deg, elevation_deg) in degree
            name       : str
        """
        super(Dipole, self).__init__(position, name)
        if len(orientation) != 2:
            raise ValueError('The length of microphone orientation should be 2 .')
        self._orient = tuple(orientation)
        self._name += "_Dipole"

    def get_orient(self):
        """
        Returns (azimuth_deg, elevation_deg) the orientation of Hypercardioid Beam Pattern 
        """
        return self._orient
    
    def __str__(self):
        return "{}_azimDeg_{:.1f}_elevDeg_{:.1f}".format(super().__str__(), *self._orient) 


class Subcardioid(Microphone):
    """
    class for Subcardioid microphone
    Args:
        position:     [x,y,z] or (x,y,z) in meter
        orientation:  (azimuth_deg, elevation_deg) in degree
        name       :  str, optional
    """
    def __init__(self, position, orientation, name=None):
        """
        Args:
            position:     [x,y,z] or (x,y,z) in meter
            orientation:  (azimuth_deg, elevation_deg) in degree
            name       : str
        """
        super(Subcardioid, self).__init__(position, name)
        if len(orientation) != 2:
            raise ValueError('The length of microphone orientation should be 2 .')
        self._orient = tuple(orientation)
        self._name += "_Subcardioid"
    
    def get_orient(self):
        """
        Returns (azimuth_deg, elevation_deg) the orientation of Hypercardioid Beam Pattern 
        """
        return self._orient
    
    def __str__(self):
        return "{}_azimDeg_{:.1f}_elevDeg_{:.1f}".format(super().__str__(), *self._orient) 





'''
pyrir Module for Simulating Room Impulse Response

author: https://github.com/ludlows
2019-11 

This program is designed with the hope that it will be useful, but WITHOUT ANY GUARANTEE.
'''
import os
import numpy as np
from scipy.io import wavfile
from .microphone import Omni, Cardioid, Dipole, Hypercardioid, Subcardioid, Microphone
from .speaker import Speaker
from .room import ReflectRoom, ReverbRoom
from .cyrir import rir


__all__ = [
    'Omni', 'Cardioid', 'Dipole', 'Hypercardioid', 'Subcardioid',
    'RIR', 'Field', 'Speaker', 'ReflectRoom', 'ReverbRoom'
    ] 


class RIR:
    """
    class for RIR (Room Impulse Response)
    Args:
        fs            : integer Sampling rate
        rir_array     : numpy array (n_channel, length of RIR)
        name          : str
        channel_names : list or tuple of str
        speaker_name  : str
    """
    _rir_id = 0
    def __init__(self, fs, rir_array, channel_names, speaker_name, name=None):
        """
        Args:
            fs            : integer Sampling rate
            rir_array     : numpy array (n_channel, length of RIR)
            name          : str
            channel_names : list or tuple of str
            speaker_name  : str
        """ 
        self._rir_array = rir_array 
        self._fs = fs
        if len(rir_array.shape) == 2:
            self._n_mic, self._n_sample = rir_array.shape
        else:
            raise ValueError('The shape of RIR Numpy Array should be 2')
        if not name:
            self._name = "RIR_{:d}_nMic_{:d}_nSample_{:d}".format(
                self._rir_id, self._n_mic, self._n_sample)
        else:
            self._name = name
        self._rir_id += 1
        self._mic_names = channel_names
        self._spk_name = speaker_name
    
    def __str__(self):
        return self._name
    
    def get_numpy(self):
        """
        Returns numpy array of RIR
        """
        return self._rir_array.copy()
    
    def apply2audio1D(self, audio1d):
        """
        Returns the reverb audio for each mirophone
        Args:
            audio1d: 1d numpy float array
        """
        length = audio1d.shape[0]
        ans = np.empty((self._n_mic, length))
        for i in range(self._n_mic):
            ans[i] = np.convolve(self._rir_array[i, :], audio1d, mode='same') 
        return ans

    def apply2audio_file(self, filepath):
        """
        Returns A list of numpy array, the length of the list is the number of auio channels
        Args:
            filepath: the audio filepath, supporting .WAV format only
        """
        fs, data = wavfile.read(filepath)
        if fs != self._fs:
            raise RuntimeError("The Sampling Rate of the Audio File is not compatible with the RIR.")
        if len(data.shape) == 1:
            n_channel = 1
        elif len(data.shape) == 2:
            n_channel = data.shape[0]
        else:
            raise ValueError('The Channel Number of Input Audio File is Wrong.')
        if not np.issubdtype(data.dtype, np.floating):
            data = data.astype(np.float32) / np.iinfo(data.dtype).max 
        if n_channel == 1:
            return [self.apply2audio1D(data)]
        return [self.apply2audio1D(data[:,i]) for i in range(n_channel)]

    def apply2audio_folder(self, audio_folder):
        """
        Returns None but writing reverb audio files with float32 format
            Supporting WAV format only for now
        Args:
           audio_foler: the clean audio folder
        """
        files = [f for f in os.listdir(audio_folder) if f.endswith('.wav') or f.endswith('.WAV')]
        out_folder = self._spk_name + "_" + "_".join(self._mic_names)
        os.makedirs(out_folder, exist_ok=True)
        for k,filename in enumerate(files):
            path = os.path.join(audio_folder, filename)
            arr_list = self.apply2audio_file(path)
            fname = ".".join(filename.split('.')[:-1])
            for i, arr in enumerate(arr_list):
                reverb_fname = fname + "_ch{:d}_Reverb.wav".format(i)
                arr32 = arr.astype(np.float32)
                wavfile.write(os.path.join(out_folder, reverb_fname), self._fs, arr32.T)
            print('Writing {:.2f}%.'.format( (k+1.0) / len(files)))
        

class Field:
    """
    class for Sound (Acoustic) Field 
    Args: 
        fs     (Hz)      : integer, Sampling Rate
        n_sample         : integer, the length of FIR (Finite Impulse Response)
        sound_speed (m/s): double,  the speed of sound
        high_pass        : bool,  enable high pass filter or not
        name             : str
    """
    _field_id = 0
    def __init__(self, fs, n_sample=1024, sound_speed=340, high_pass=True, name=None):
        """
        Args: 
            fs     (Hz)      : integer, Sampling Rate
            n_sample         : integer, the length of FIR (Finite Impulse Response)
            sound_speed (m/s): double,  the speed of sound
            high_pass        : bool,  enable high pass filter or not
            name             : str
        """
        self._fs = fs 
        self._n_sample = n_sample
        self._sound_speed = sound_speed
        if not name:
            self._name = "Field_{:d}_fs_{:d}_nSample_{:d}".format(self._field_id, fs, n_sample)
        else:
            self._name = name
        self._field_id += 1
        self._high_pass = high_pass
    
    def compute_rir(self, room):
        """
        Args:
            room: Room Object with Microphoena and Speaker
        Returns tuple of RIR objects
        """
        # beta array
        if isinstance(room, ReflectRoom):
            beta = room.get_beta_array()
        else: # reverb Room
            # calculate beta array
            rt60 = room.get_rt60()
            x, y, z  = room.get_size()
            vol = x * y * z
            area = 2 * (x*y + x*z + y*z)
            alpha = 24.0 * vol * np.log(10.0) / (self._sound_speed * area * rt60)
            if alpha > 1:
                raise ValueError("The Room Size makes the wall reflection coefficients invalid.")
            beta0 = np.sqrt(1.0-alpha)
            beta = tuple(beta0 for _ in range(6))
        beta = np.array(beta, dtype=np.float64)

        comb = room.mic_speaker_combination()
        n_mic = len(comb[0][0])
        room_size = np.array(room.get_size(), dtype=np.float64)
        rirobj_list = []
        for mic_arr, spk in comb:
            src_pos = np.array(spk.get_pos(),  dtype=np.float64)
            mic_names = [str(mic) for mic in mic_arr]
            rir_numpy_mics = np.empty((n_mic, self._n_sample), dtype=np.float64)
            for i, mic in enumerate(mic_arr):
                mic_pos = np.array(mic.get_pos(), dtype=np.float64)
                if isinstance(mic, Dipole):
                    mic_type = ord('d')
                    mic_azimuth, mic_elevation = mic.get_orient()
                elif isinstance(mic, Subcardioid):
                    mic_type = ord('s')
                    mic_azimuth, mic_elevation = mic.get_orient()
                elif isinstance(mic, Cardioid):
                    mic_type = ord('c')
                    mic_azimuth, mic_elevation = mic.get_orient()
                elif isinstance(mic, Hypercardioid):
                    mic_type = ord('h')
                    mic_azimuth, mic_elevation = mic.get_orient()
                else: # Omni
                    mic_type = ord('o')
                    mic_azimuth, mic_elevation = (0.0,0.0)
                mic_azimuth   = mic_azimuth / 180.0 * np.pi
                mic_elevation = mic_elevation / 180.0 * np.pi 
                rir_numpy = rir(
                    self._sound_speed,
                    self._fs,
                    room_size,
                    mic_pos,
                    src_pos,
                    beta,
                    mic_azimuth,
                    mic_elevation,
                    self._n_sample,
                    self._high_pass,
                    mic_type,
                    room.get_reflect_order()
                )
                rir_numpy_mics[i] = rir_numpy
            
            rirobj_list.append(RIR(self._fs, rir_numpy_mics, tuple(mic_names), str(spk)))
        return tuple(rirobj_list)
    
    def __str__(self):
        return self._name


# 2019-12
# github.com/ludlows
# Lite Python Package for Room Impulse Response
import numpy
from setuptools import find_packages
from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import build_ext,cythonize

with open("README.md", "r") as fh:
    long_description = fh.read()

extensions = [
    Extension(
        "pyrir.cyrir.cyrir",
        ["pyrir/cyrir/cyrir.pyx"],
        include_dirs=[numpy.get_include()],
        language="c")
]
setup(
    name="pyrir",
    version="0.0.1",
    author="ludlows",
    description="Lite Package for Room Impulse Response",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ludlows/pyrir",
    packages=find_packages(),
    package_data={'pyrir':["cyrir/cyrir.pyx", "cyrir/rir.c"]},
    cmdclass = {'build_ext': build_ext},
    ext_modules=cythonize(extensions),
    setup_requires=['numpy', 'scipy', 'cython'],
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)